from typing import Any

from pydantic import BaseModel

from lythonic import Result
from lythonic.compose.logic import LogicNode


def test_logic_node():
    def m1(config: BaseModel) -> BaseModel:
        return config

    edge_method = LogicNode(m1)
    assert edge_method.input_types == [BaseModel]
    assert edge_method.ok_output_types == [BaseModel]
    assert edge_method.err_output_type is None

    def m2(config: BaseModel) -> tuple[BaseModel, BaseModel]:
        return config, BaseModel()

    edge_method = LogicNode(m2)
    assert edge_method.input_types == [BaseModel]
    assert edge_method.ok_output_types == [BaseModel, BaseModel]
    assert edge_method.err_output_type is None

    def m3(config: BaseModel) -> Result[BaseModel, Any]:
        return Result[BaseModel, Any].Ok(config)

    edge_method = LogicNode(m3)
    assert edge_method.input_types == [BaseModel]
    assert edge_method.ok_output_types == [BaseModel]
    assert edge_method.err_output_type is not None and edge_method.err_output_type.types == [Any]


# import asyncio
# from collections.abc import Callable
# from typing import Any, final

# from lythonic import GlobalRef

# from lythonic.logic import Logic

# def assert_ae(
#     call: Callable[[], Any],
#     expected_to_start_with: str,
#     expected_exception: Any = AssertionError,
# ):
#     try:
#         call()
#     except Exception as e:
#         print(repr(str(e)))
#         assert isinstance(e, expected_exception)
#         assert str(e).startswith(expected_to_start_with)
#     else:
#         raise AssertionError(f"Expected AssertionError({expected_to_start_with!r}) not raised")


# def test_logic():
#     print(GlobalRef(test_logic))
#     assert Logic({"ref$": "test_gref:s2"}).call() == 2
#     assert_ae(lambda: Logic({"ref$": "test_gref:s2", "a": 3}), "Unexpected entries {'a': 3}")
#     assert asyncio.run(Logic({"ref$": "test_gref:A2"}).call(2)) == 4
#     assert asyncio.run(Logic({"ref$": "test_gref:A2", "a": 5}).call(2)) == 7
#     assert_ae(
#         lambda: Logic({"ref$": "test_gref:A2", "x": 5}),
#         "not supported keys in config: {'x': 5}",
#     )
#     assert Logic({"ref$": "test_gref:M3"}).call(2) == 6
#     assert Logic({"ref$": "test_gref:M3", "i": 5}).call(2) == 10
#     assert_ae(
#         lambda: Logic({"ref$": "test_gref:A2", "x": 4}),
#         "not supported keys in config: {'x': 4}",
#     )
#     assert Logic({"i": 5}, "test_gref:M3").call(2) == 10
#     assert Logic({"ref$": "test_gref:M3", "i": 5}, "test_gref:M4").call(2) == 10
#     assert_ae(
#         lambda: Logic({"ref$": "test_gref:M4", "i": 5}, "test_gref:M3").call(2),
#         "module 'test_gref' has no attribute 'M4'",
#         AttributeError,
#     )



"""
Compose: Build typed callable compositions from annotated functions.

This module provides introspection and composition primitives for building
higher-level structures (CLIs, pipelines, workflows) from type-annotated
callables.

## Core Concepts

- `Method`: Wrapper around a callable with signature introspection
- `MethodDict`: Dictionary of methods indexed by name
- `ArgInfo`: Metadata about function arguments

## CLI Support

For building command-line interfaces, see `lythonic.compose.cli`:

```python
from lythonic.compose.cli import ActionTree, Main, RunContext
```
"""

from __future__ import annotations

import inspect
from collections.abc import Callable
from typing import Any, Generic, NamedTuple, TypeVar

from pydantic import BaseModel

from lythonic import GlobalRef


class ArgInfo(NamedTuple):
    """
    Metadata about a function argument, extracted from signature and Pydantic fields.

    Used to generate CLI argument/option parsing and help text.
    """

    name: str
    annotation: Any | None
    default: Any | None
    is_optional: bool
    description: str

    @classmethod
    def from_param(cls, param: inspect.Parameter, origin: Any):
        description = ""
        is_optional = param.default != inspect.Parameter.empty
        default = param.default if is_optional else None
        if isinstance(origin, type) and issubclass(origin, BaseModel):
            if param.name in origin.model_fields:
                field = origin.model_fields[param.name]
                if field.description is not None:
                    description = field.description
                is_optional = not field.is_required()
                default = field.default
        return cls(
            name=param.name,
            annotation=param.annotation if param.annotation != inspect.Parameter.empty else None,
            default=default,
            is_optional=is_optional,
            description=description,
        )

    def to_value(self, v: str):
        if self.annotation is None:
            return v
        if self.annotation is bool:
            return v.lower() in ("true", "1", "yes", "y")
        if issubclass(self.annotation, BaseModel):
            return self.annotation.model_validate_json(v)
        return self.annotation(v)

    def is_turn_on_option(self) -> bool:
        return self.annotation is bool and self.default is False

    @property
    def type(self) -> str:
        if self.annotation is not None:
            return self.annotation.__name__
        return "str"

    def arg_help(self, indent: int):
        return f"{' ' * indent}<{self.name}> - {self.type}: {self.description}"

    def opt_help(self, indent: int):
        return f"{' ' * indent}[--{self.name}{'=value' if not self.is_turn_on_option() else ''}] - {self.type}: {self.description}. Default: {self.default!r}"


class Method:
    """
    Wrapper around a callable that provides introspection of its arguments.

    Lazily loads the callable via GlobalRef and extracts argument metadata
    from the function signature. Supports both regular functions and Pydantic
    BaseModel classes (using their `__init__` signature).
    """

    gref: GlobalRef
    _o: Callable[..., Any] | None
    _args: list[ArgInfo] | None
    _args_by_name: dict[str, ArgInfo] | None
    _return_annotation: Any | None

    def __init__(self, o: Callable[..., Any] | GlobalRef):
        if isinstance(o, GlobalRef):
            self.gref = o
            self._o = None
        else:
            self.gref = GlobalRef(o)
            assert isinstance(o, Callable), "method instance must be a callable"
            self._o = o
        self._args = None
        self._args_by_name = None
        self._return_annotation = None

    def _update_from_signature(self):
        o = self.o
        sig = inspect.signature(o)
        self._args = [ArgInfo.from_param(param, origin=o) for param in sig.parameters.values()]
        self._args_by_name = {arg.name: arg for arg in self._args}
        self._return_annotation = sig.return_annotation

    @property
    def o(self) -> Callable[..., Any]:
        if self._o is None:
            self._o = self.gref.get_instance()
        assert self._o is not None
        return self._o

    @property
    def args(self) -> list[ArgInfo]:
        if self._args is None:
            self._update_from_signature()
        assert self._args is not None
        return self._args

    @property
    def args_by_name(self) -> dict[str, ArgInfo]:
        if self._args_by_name is None:
            self._update_from_signature()
        assert self._args_by_name is not None
        return self._args_by_name

    @property
    def return_annotation(self) -> Any | None:
        if self._args is None:
            self._update_from_signature()
        return self._return_annotation

    @property
    def name(self):
        return self.gref.name

    @property
    def doc(self):
        return self.o.__doc__

    def __call__(self, *args: Any, **kwargs: Any):
        return self.o(*args, **kwargs)


T = TypeVar("T", bound=Method)


class MethodDict(Generic[T], dict[str, T]):
    """
    Dictionary mapping lowercased method names to Method instances.

    Use `add()` to register a callable, or `wrap()` as a decorator.
    """

    method_type: type[T]

    def __init__(self, method_type: type[T]):
        super().__init__()
        self.method_type = method_type

    def add(self, o: Callable[..., Any]) -> T:
        m = self.method_type(o)
        self[m.name.lower()] = m
        return m

    def wrap(self, o: Callable[..., Any]) -> T:
        return self.add(o)



"""
CLI: Build hierarchical command-line interfaces from type-annotated functions.

This module provides a declarative way to build CLI applications where commands
are organized in a tree structure. Arguments and options are automatically derived
from function signatures and Pydantic models.

## Quick Start

```python
from pydantic import BaseModel, Field
from lythonic.compose.cli import ActionTree, Main, RunContext

# 1. Create the root action tree
main = ActionTree(Main)

# 2. Define a subcommand group using a BaseModel
class Server(BaseModel):
    '''Server management commands'''
    port: int = Field(default=8080, description="Port to listen on")

server = main.actions.add(Server)

# 3. Add actions to the group using the @wrap decorator
@server.actions.wrap
def start(ctx: RunContext):
    '''Start the server'''
    config = ctx.path.get('/server')  # Access parent action's parsed values
    ctx.print(f"Starting on port {config.port}")

@server.actions.wrap
def stop():
    '''Stop the server'''
    print("Stopping server")

# 4. Run the CLI
if __name__ == "__main__":
    import sys
    result = main.run_args(sys.argv)
    sys.exit(0 if result.success else 1)
```

Usage: `mycli server start` or `mycli server --port=9000 start`

## Key Concepts

- **ActionTree**: A node in the command tree. Can have child actions and arguments.
- **Main**: Default root model with `--help` flag. Use as the root ActionTree argument.
- **RunContext**: Passed to actions that need it (declare `ctx: RunContext` as first param).
  Provides access to parent values via `ctx.path.get('/path')`.
- **Arguments**: Required params become positional args; optional params become `--flags`.
- **Pydantic integration**: BaseModel fields provide descriptions and defaults automatically.

## Argument Parsing Rules

- Required function params → positional arguments: `<arg>`
- Optional function params → options: `--name=value` or `--name value`
- Boolean params with `default=False` → flags: `--verbose` (no value needed)
- Pydantic models can be passed as JSON: `mycli config set '{"name":"test"}'`
"""

from __future__ import annotations

import sys
from collections.abc import Callable
from pathlib import Path
from typing import Any

from pydantic import BaseModel, Field
from typing_extensions import override

from lythonic import GlobalRef
from lythonic.compose import ArgInfo, Method, MethodDict


class RunResult:
    """Result of running a CLI command, with success status and collected messages."""

    success: bool
    msgs: list[str]
    print_func: Callable[[str], None] | None

    def __init__(self, success: bool = False, print_func: Callable[[str], None] | None = print):
        self.success = success
        self.msgs = []
        self.print_func = print_func

    def print(self, msg: str):
        if self.print_func is None:
            self.msgs.append(msg)
        else:
            self.print_func(msg)


class ActionTree(Method):
    """
    A node in the command tree that can have child actions.

    Extends Method with the ability to register subcommands via the `actions`
    MethodDict. Call `run_args(sys.argv)` on the root to parse and execute.
    """

    actions: MethodDict[ActionTree]

    def __init__(self, o: Callable[..., Any] | GlobalRef):
        super().__init__(o)
        self.actions = MethodDict[ActionTree](method_type=self.__class__)

    def run_args(
        self, argv: list[str], print_func: Callable[[str], None] | None = print
    ) -> RunResult:
        cli_name = Path(argv[0]).name
        cli_args = argv[1:]
        return self._run_args(cli_args, RunContext(self, cli_name, print_func))

    def _split_ctx_args_opts(self) -> tuple[bool, list[ArgInfo], list[ArgInfo]]:
        has_ctx: bool = (
            len(self.args) > 0
            and self.args[0].name == "ctx"
            and self.args[0].annotation == RunContext
        )
        self_args = self.args[1:] if has_ctx else self.args
        return (
            has_ctx,
            [arg for arg in self_args if not arg.is_optional],
            [arg for arg in self_args if arg.is_optional],
        )

    def _run_args(self, cli_args: list[str], ctx: RunContext) -> RunResult:
        current_arg_index = 0
        arg_values = {}
        has_ctx, required_args, _ = self._split_ctx_args_opts()
        if has_ctx:
            arg_values["ctx"] = ctx
        error = None
        try:
            while current_arg_index < len(cli_args):
                arg_str = cli_args[current_arg_index]
                if required_args:
                    arg = required_args[0]
                    if arg_str.startswith("--"):
                        raise ValueError(
                            f"Argument {arg.name} is required, but getting option {arg_str}"
                        )
                    current_arg_index += 1
                    arg_values[arg.name] = arg.to_value(arg_str)
                    required_args.pop(0)
                    continue
                if arg_str.startswith("--"):
                    arg_str = arg_str[2:]
                    if "=" in arg_str:
                        k, v = arg_str.split("=", 1)
                    else:
                        k = arg_str
                        v = None
                    if k in arg_values:
                        raise ValueError(f"--{k} is already set to {arg_values[k]}")
                    if k not in self.args_by_name:
                        raise ValueError(
                            f"--{k} is not a valid option, expected one of {list(map(lambda x: f'--{x}', self.args_by_name.keys()))}"
                        )
                    arg = self.args_by_name[k]
                    if v is None:
                        if arg.is_turn_on_option():
                            v = "y"
                        else:
                            current_arg_index += 1
                            if current_arg_index >= len(cli_args):
                                raise ValueError(f"Value is not provided for --{k}")
                            v = cli_args[current_arg_index]
                    arg_values[k] = arg.to_value(v)
                    current_arg_index += 1
                    continue
                elif arg_str in self.actions:
                    action: ActionTree = self.actions[arg_str]
                    ctx.path.value = self(**arg_values)
                    ctx.add_path(arg_str)
                    return action._run_args(cli_args[current_arg_index + 1 :], ctx)
                raise ValueError(f"Argument {arg_str!r} is not a valid")
            if required_args:
                raise ValueError(
                    f"Required arguments are missing: {' '.join(f'<{arg.name}>' for arg in required_args)}"
                )
            if self.actions:
                raise ValueError(
                    f"Action need to be specified, expected one of {', '.join(self.actions.keys())}"
                )
        except Exception:
            error = str(sys.exc_info()[1])
        if error or ctx.is_print_help_selected():
            ctx.print_help(error, ctx.path, self)
        else:
            ctx.path.value = self(**arg_values)
            ctx.run_result.success = True
        return ctx.run_result


class Main(BaseModel):
    """Default root action model. Provides the `--help` flag."""

    help: bool = Field(default=False, description="Show help")


class PathValue:
    """
    Records values along a path.

    >>> root = PathValue.root()
    >>> root.value = "I am Groot"
    >>> foo = PathValue("foo", root, "I am foo")
    >>> foo.get("/foo")
    'I am foo'
    >>> foo.parts
    ('', 'foo')
    >>> root.parts
    ('',)
    >>> foo.get("/bar")
    Traceback (most recent call last):
    ...
    ValueError: Path /bar is not contained in /foo
    >>> foo.get("/foo/")
    'I am foo'
    >>> foo.get("/")
    'I am Groot'


    """

    name: str
    parent: PathValue | None
    value: Any | None

    def __init__(self, name: str, parent: PathValue | None, value: Any | None = None) -> None:
        if name == "":
            assert parent is None
        else:
            assert parent is not None
        self.name = name
        self.parent = parent
        self.value = value

    @classmethod
    def root(cls) -> PathValue:
        return cls(name="", parent=None)

    @property
    def parts(self) -> tuple[str, ...]:
        if self.parent is None:
            return (self.name,)
        return (*self.parent.parts, self.name)

    def get_back(self, n: int) -> Any | None:
        if n == 0:
            return self.value
        if self.parent is None:
            return None
        return self.parent.get_back(n - 1)

    def get(self, path: str) -> Any | None:
        parts = self.parts
        lookup_parts = tuple(path.split("/"))
        if lookup_parts[-1] == "":
            lookup_parts = lookup_parts[:-1]
        lpl = len(lookup_parts)
        pl = len(parts)
        if parts[:lpl] != lookup_parts:
            raise ValueError(f"Path {path} is not contained in {self}")
        return self.get_back(pl - lpl)

    @override
    def __str__(self):
        return "/".join(self.parts)


class RunContext:
    """
    Context passed to action functions during CLI execution.

    Declare `ctx: RunContext` as the first parameter of an action to receive it.
    Use `ctx.path.get('/action_name')` to access parsed values from parent actions.
    Use `ctx.print()` for output that gets captured in `RunResult.msgs`.
    """

    path: PathValue
    main_at: ActionTree
    run_result: RunResult
    cli_name: str

    def __init__(
        self, main_at: ActionTree, cli_name: str, print_func: Callable[[str], None] | None = print
    ):
        self.path = PathValue.root()
        self.main_at = main_at
        self.run_result = RunResult(print_func=print_func)
        self.cli_name = cli_name

    def add_path(self, path: str):
        self.path = PathValue(path, self.path)

    def is_print_help_selected(self) -> bool:
        v = self.path.get("/")
        if isinstance(v, Main):
            return v.help
        return False

    def print(self, msg: str):
        self.run_result.print(msg)

    def _recursive_help(self, print_at: ActionTree, indent: int):
        indent += 2
        _, arguments, options = print_at._split_ctx_args_opts()  # pyright: ignore[reportPrivateUsage]
        for arg in arguments:
            self.print(arg.arg_help(indent))
        for opt in options:
            self.print(opt.opt_help(indent))
        if print_at.actions:
            self.print(f"{' ' * indent}Actions:")
            indent += 2
            for action_name, action in print_at.actions.items():
                self.print(f"{' ' * indent}{action_name} - {action.doc}")
                self._recursive_help(action, indent + 2)

    def print_help(self, error: str | None, path: PathValue, current_at: ActionTree):
        if error:
            self.print(f"Error: {error}")
        self.print("Usage: ")
        self.print(f"  {self.cli_name} {' '.join(path.parts[1:])}")
        self._recursive_help(current_at, 2)



import difflib
import re
from pathlib import Path

from pydantic import BaseModel, Field

from lythonic.compose import Method
from lythonic.compose.cli import ActionTree, Main, RunContext


def my_func(x, y: bool, z=None, a: int | float = 42, b: str = "foo"):  # pyright: ignore
    pass


class Config(BaseModel):
    name: str = Field(description="The name of the server")


def zet(config: Config):
    """Set the server configuration"""
    print(config)


def test_method():
    z = Method(zet)
    assert len(z.args) == 1
    assert z.args[0].name == "config"
    assert z.args[0].annotation is Config
    assert z.args[0].default is None
    assert z.args[0].is_optional is False
    assert z.args[0].description == ""
    assert z.args[0].arg_help(0) == "<config> - Config: "

    main = Method(Main)
    assert len(main.args) == 1
    assert main.args[0].name == "help"
    assert main.args[0].annotation is bool
    assert main.args[0].default is False
    assert main.args[0].is_optional is True
    assert main.args[0].description == "Show help"
    assert main.args[0].opt_help(0) == "[--help] - bool: Show help. Default: False"

    m = Method(my_func)  # pyright: ignore
    assert len(m.args) == 5
    assert m.args[0].name == "x"
    assert m.args[0].annotation is None
    assert m.args[0].default is None
    assert m.args[0].is_optional is False
    assert m.args[0].description == ""

    assert m.args[1].name == "y"
    assert m.args[1].annotation is bool
    assert m.args[1].default is None
    assert m.args[1].is_optional is False
    assert m.args[1].description == ""

    assert m.args[2].name == "z"
    assert m.args[2].annotation is None
    assert m.args[2].default is None
    assert m.args[2].is_optional is True
    assert m.args[2].description == ""

    assert m.args[3].name == "a"
    assert str(m.args[3].annotation) == "int | float"
    assert m.args[3].default == 42
    assert m.args[3].is_optional is True
    assert m.args[3].description == ""  # "an integer or a float"

    assert m.args[4].name == "b"
    assert m.args[4].annotation is str
    assert m.args[4].default == "foo"
    assert m.args[4].is_optional is True
    assert m.args[4].description == ""


main_at = ActionTree(Main)


class Server(BaseModel):
    """Managing the server"""

    data: Path = Field(default=Path("./data"), description="directory to store all server data")


server_at = main_at.actions.add(Server)


@server_at.actions.wrap
def start(ctx: RunContext):
    """Starts the server in the foreground"""
    ctx.print(f"Starting server with data directory: {ctx.path.get('/server').data}")  # pyright: ignore[reportOptionalMemberAccess]


@server_at.actions.wrap
def stop():
    """Stops the server"""
    print("stopping server ")


@server_at.actions.wrap
def config():
    """Manage the server configuration"""
    pass


@config.actions.wrap
def set(config: Config):
    """Set the server configuration"""
    print(config)


@config.actions.wrap
def get():
    """Get the server configuration"""
    print("config get")


def test_at():
    has_ctx, args, opts = set._split_ctx_args_opts()  # pyright: ignore[reportPrivateUsage]
    assert has_ctx is False
    assert len(opts) == 0
    assert len(args) == 1


def test_cli():
    def run(cmd: str, success: bool, expected: list[str]):
        rr = main_at.run_args(cmd.split(), print_func=None)
        assert rr.success == success

        def match_output(actual: list[str], expected: list[str]) -> bool:
            if len(actual) != len(expected):
                return False
            for a, e in zip(actual, expected, strict=True):
                if e[:1] == "~":
                    if not re.compile(e[1:]).match(a.strip()):
                        return False
                elif a.strip() != e.strip():
                    return False
            return True

        match = match_output(rr.msgs, expected)
        if not match:
            print(rr.msgs)
            print("\n".join(difflib.unified_diff(rr.msgs, expected)))

        return match

    path_type = type(Path("/")).__name__
    full_help = [
        "Usage: ",
        "  wgl ",
        "    [--help] - bool: Show help. Default: False",
        "    Actions:",
        "      server - Managing the server",
        f"          [--data=value] - Path: directory to store all server data. Default: {path_type}('data')",
        "          Actions:",
        "            start - Starts the server in the foreground",
        "            stop - Stops the server",
        "            config - Manage the server configuration",
        "                Actions:",
        "                  set - Set the server configuration",
        "                      <config> - Config: ",
        "                  get - Get the server configuration",
    ]
    assert run("wgl -help", False, ["Error: Argument '-help' is not a valid", *full_help])
    assert run(
        "wgl --help",
        False,
        ["Error: Action need to be specified, expected one of server", *full_help],
    )
    assert run(
        "wgl --help=y",
        False,
        ["Error: Action need to be specified, expected one of server", *full_help],
    )
    assert run("wgl --help=y --help", False, ["Error: --help is already set to True", *full_help])
    assert run(
        "wgl server",
        False,
        [
            "Error: Action need to be specified, expected one of start, stop, config",
            "Usage: ",
            "  wgl server",
            *full_help[5:],
        ],
    )
    assert run("wgl server start", True, ["Starting server with data directory: data"])
    assert run("wgl server stop", True, [])
    assert run(
        "wgl server config",
        False,
        [
            "Error: Action need to be specified, expected one of set, get",
            "Usage: ",
            "  wgl server config",
            *full_help[10:],
        ],
    )
    assert run(
        "wgl server config set",
        False,
        [
            "Error: Required arguments are missing: <config>",
            "Usage: ",
            "  wgl server config set",
            "    <config> - Config: ",
        ],
    )
    assert run("wgl server config get", True, [])
    assert run(
        "wgl server config set {}",
        False,
        [
            r"~Error: 1 validation error for Config\s+name\s+Field required .*",
            "Usage: ",
            "  wgl server config set",
            "    <config> - Config: ",
        ],
    )
    assert run('wgl server config set {"name":"test"}', True, [])



"""
Time and scheduling utilities.

This module provides tools for working with time intervals, frequencies, and periodic tasks:

## Time Simulation

`SimulatedTime` allows testing time-dependent code by offsetting the clock:

```python
from lythonic.periodic import stime
from datetime import timedelta

stime.set_offset(timedelta(days=1))  # Pretend it's tomorrow
print(stime.get_datetime())
stime.reset()  # Back to real time
```

## Frequencies and Intervals

- `Frequency`: Human-friendly periods (weekly, monthly, quarterly, annually)
- `FrequencyOffset`: Frequency with day offset (e.g., "20th of each month")
- `Interval`: Precise duration with multiplier (e.g., "3M" for 3 months, "2W" for 2 weeks)

```python
from lythonic.periodic import Frequency, Interval
from datetime import date

freq = Frequency("monthly")
print(freq.first_day(date(2025, 11, 15)))  # 2025-11-01
print(freq.last_day(date(2025, 11, 15)))   # 2025-11-30

interval = Interval.from_string("2W")
print(interval.timedelta())  # 14 days
```

## Periodic Tasks

Run async tasks at specified intervals:

```python
from lythonic.periodic import PeriodicTask, run_all

task = PeriodicTask(freq=60, logic=my_function)  # Run every 60 seconds
await run_all(task)
```

## Timing with Moment

Track elapsed time between checkpoints:

```python
from lythonic.periodic import Moment

m = Moment.start()
# ... do work ...
m = m.capture("step1")
# ... more work ...
m = m.capture("done")
print(m.chain())  # [start] 0.5s-> [step1] 1.2s-> [done]
```
"""

import asyncio
import calendar
import inspect
import logging
import sys
import time as tt
from collections.abc import Callable
from typing import Annotated, Any, Literal, final

from pydantic import BeforeValidator, PlainSerializer, WithJsonSchema
from typing_extensions import override

from lythonic import str_or_none, utc_now
from lythonic.types import KNOWN_TYPES, KnownTypeArgs

log = logging.getLogger(__name__)

import re
from datetime import UTC, date, datetime, timedelta
from enum import Enum
from pathlib import Path

YEAR_IN_DAYS = 365.256
SECONDS_IN_DAY = 24 * 60 * 60


EPOCH_ZERO = datetime(1970, 1, 1, tzinfo=UTC)


def total_microseconds(d: timedelta) -> int:
    return (d.days * SECONDS_IN_DAY + d.seconds) * 1_000_000 + d.microseconds


def dt_to_bytes(dt: datetime) -> bytes:
    """
    Convert datetime to bytes

    >>> dt_to_bytes(datetime( 1900,1,1,0,0,0))
    b'\\xff\\xf8&\\xef\\xb7C`\\x00'
    >>> dt_to_bytes(datetime( 2000,1,1,0,0,0))
    b'\\x00\\x03]\\x01;7\\xe0\\x00'
    """
    if dt.tzinfo is None:
        dt = dt.replace(tzinfo=UTC)
    mics = total_microseconds(dt - EPOCH_ZERO)
    return mics.to_bytes(8, "big", signed=True)


def dt_from_bytes(b: bytes) -> datetime:
    """
    Convert  bytes to datetime

    >>> dt_from_bytes(b'\\xff\\xf8&\\xef\\xb7C`\\x00')
    datetime.datetime(1900, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)
    >>> dt_from_bytes(b'\\x00\\x03]\\x01;7\\xe0\\x00')
    datetime.datetime(2000, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)
    """
    mics = int.from_bytes(b, "big", signed=True)
    return EPOCH_ZERO + timedelta(microseconds=mics)


DT_BYTES_LENGTH = len(dt_to_bytes(utc_now()))


class SimulatedTime:
    """
    >>> st = SimulatedTime()
    >>> st.get_datetime().tzinfo
    datetime.timezone.utc
    >>> cmp = lambda ss: abs((st.get_datetime()-utc_now()).total_seconds()-ss) < 1e-3
    >>> cmp(0)
    True
    >>> st.set_offset(timedelta(days=1))
    >>> cmp(86400)
    True
    >>> st.set_offset(timedelta(days=1).total_seconds())
    >>> cmp(86400)
    True
    >>> st.set_now(utc_now() + timedelta(days=1))
    >>> cmp(86400)
    True
    >>> st.set_now( (utc_now() - timedelta(days=1)).timestamp() )
    >>> cmp(-86400)
    True
    >>> st.is_real_time()
    False
    >>> st.reset()
    >>> st.is_real_time()
    True
    """

    offset: float

    def __init__(self, offset: float = 0.0) -> None:
        self.offset = offset

    def time(self):
        return tt.time() + self.offset

    def set_offset(self, offset: timedelta | float):
        if isinstance(offset, timedelta):
            self.offset = offset.total_seconds()
        else:
            self.offset = offset

    def set_now(self, dt: datetime | float):
        if isinstance(dt, datetime):
            epoch = dt.timestamp()
        else:
            epoch = dt
        self.offset = epoch - tt.time()

    def reset(self):
        self.offset = 0.0

    def is_real_time(self):
        return self.offset == 0.0

    def get_datetime(self) -> datetime:
        return datetime.fromtimestamp(self.time(), tz=UTC)


stime: SimulatedTime = SimulatedTime()

FrequencyType = Literal["weekly", "monthly", "quarterly", "annually"]


class Frequency:
    """People friendly interval

    >>> Frequency("weekly").first_day(date(2025, 11, 21))
    datetime.date(2025, 11, 17)
    >>> Frequency("weekly").last_day(date(2025, 11, 21))
    datetime.date(2025, 11, 23)
    >>> Frequency("monthly").first_day(date(2025, 11, 21))
    datetime.date(2025, 11, 1)
    >>> Frequency("monthly").last_day(date(2025, 11, 21))
    datetime.date(2025, 11, 30)
    >>> Frequency("quarterly").first_day(date(2025, 11, 21))
    datetime.date(2025, 10, 1)
    >>> Frequency("quarterly").last_day(date(2025, 11, 21))
    datetime.date(2025, 12, 31)
    >>> Frequency("quarterly").first_day(date(2025, 2, 21))
    datetime.date(2025, 1, 1)
    >>> Frequency("quarterly").last_day(date(2025, 2, 21))
    datetime.date(2025, 3, 31)
    >>> Frequency("quarterly").last_day(date(2025, 5, 21))
    datetime.date(2025, 6, 30)
    >>> Frequency("annually").first_day(date(2025, 11, 21))
    datetime.date(2025, 1, 1)
    >>> Frequency("annually").last_day(date(2025, 11, 21))
    datetime.date(2025, 12, 31)
    """

    frequency: FrequencyType

    def __init__(self, frequency: FrequencyType) -> None:
        self.frequency = frequency

    @classmethod
    def ensure(cls, frequency: "Frequency|FrequencyType") -> "Frequency":
        if isinstance(frequency, Frequency):
            return frequency
        return cls(frequency)

    def first_day(self, as_of: date) -> date:
        if self.frequency == "weekly":
            return as_of - timedelta(days=as_of.weekday())
        elif self.frequency == "monthly":
            return as_of.replace(day=1)
        elif self.frequency == "quarterly":
            return as_of.replace(day=1, month=((as_of.month - 1) // 3) * 3 + 1)
        elif self.frequency == "annually":
            return as_of.replace(day=1, month=1)
        else:
            raise AssertionError(f"Invalid frequency: {self.frequency}")

    def last_day(self, as_of: date) -> date:
        if self.frequency == "weekly":
            return as_of + timedelta(days=6 - as_of.weekday())
        elif self.frequency == "monthly":
            return as_of.replace(day=calendar.monthrange(as_of.year, as_of.month)[1])
        elif self.frequency == "quarterly":
            month = ((as_of.month - 1) // 3) * 3 + 3
            return as_of.replace(day=calendar.monthrange(as_of.year, month)[1], month=month)
        elif self.frequency == "annually":
            return as_of.replace(day=31, month=12)
        else:
            raise AssertionError(f"Invalid frequency: {self.frequency}")

    def min_max_offset(
        self,
    ) -> tuple[int, int]:
        if self.frequency == "weekly":
            return (-7, 7)
        elif self.frequency == "monthly":
            return (-31, 31)
        elif self.frequency == "quarterly":
            return (-92, 92)
        elif self.frequency == "annually":
            return (-366, 366)
        else:
            raise AssertionError(f"Invalid frequency: {self.frequency}")

    def assert_offset(self, offset: int) -> None:
        min_offset, max_offset = self.min_max_offset()
        assert min_offset <= offset < max_offset, (
            f"Invalid offset: {offset} for frequency: {self.frequency} expected between {min_offset} and {max_offset}"
        )


class FrequencyOffset:
    """
    >>> FrequencyOffset(Frequency("weekly"), 0).boundaries(date(2025, 11, 21))
    (datetime.date(2025, 11, 17), datetime.date(2025, 11, 23))
    >>> FrequencyOffset("weekly", 1).boundaries(date(2025, 11, 21))
    (datetime.date(2025, 11, 18), datetime.date(2025, 11, 24))
    >>> FrequencyOffset(Frequency("weekly"), -1).boundaries(date(2025, 11, 21))
    (datetime.date(2025, 11, 16), datetime.date(2025, 11, 22))
    >>> on20thOfMonth = FrequencyOffset(Frequency("monthly"), 19)
    >>> on20thOfMonth.boundaries(date(2025, 11, 21))
    (datetime.date(2025, 11, 20), datetime.date(2025, 12, 19))
    >>> on20thOfMonth.boundaries(on20thOfMonth.boundaries(date(2025, 11, 21))[1]+timedelta(days=1))
    (datetime.date(2025, 12, 20), datetime.date(2026, 1, 19))
    >>> on3rdBeforeEndOfMonth = FrequencyOffset("monthly", -3)
    >>> b1 = on3rdBeforeEndOfMonth.boundaries(date(2025, 11, 21)); b1
    (datetime.date(2025, 10, 29), datetime.date(2025, 11, 27))
    >>> b2 = on3rdBeforeEndOfMonth.boundaries(b1[1]+timedelta(days=1)); b2
    (datetime.date(2025, 11, 28), datetime.date(2025, 12, 28))
    >>> b3 = on3rdBeforeEndOfMonth.boundaries(b2[1]+timedelta(days=1)); b3
    (datetime.date(2025, 12, 29), datetime.date(2026, 1, 28))
    >>> b4 = on3rdBeforeEndOfMonth.boundaries(b3[1]+timedelta(days=1)); b4
    (datetime.date(2026, 1, 29), datetime.date(2026, 2, 25))
    """

    frequency: Frequency
    offset: int

    def __init__(self, frequency: Frequency | FrequencyType, offset: int) -> None:
        self.frequency = Frequency.ensure(frequency)
        self.frequency.assert_offset(offset)
        self.offset = offset

    def boundaries(self, as_of: date) -> tuple[date, date]:
        """
        Return the boundaries around the given date (inclusive).
        """
        if self.offset >= 0:
            d = self.frequency.first_day(as_of)
            dd = (
                self.frequency.first_day(d - timedelta(days=1)),
                d,
                self.frequency.last_day(as_of) + timedelta(days=1),
            )
            with_delta = [t + timedelta(days=self.offset) for t in dd]
        else:
            d = self.frequency.last_day(as_of)
            dd = (
                self.frequency.first_day(as_of) - timedelta(days=1),
                d,
                self.frequency.last_day(d + timedelta(days=1)),
            )
            with_delta = [t + timedelta(days=1 + self.offset) for t in dd]

        assert with_delta[0] <= as_of
        if as_of < with_delta[1]:
            return with_delta[0], with_delta[1] - timedelta(days=1)
        else:
            assert with_delta[1] <= as_of < with_delta[2]
            return with_delta[1], with_delta[2] - timedelta(days=1)


class IntervalUnit(Enum):
    """
    >>> IntervalUnit.D
    IntervalUnit.D
    """

    D = 1
    W = 7
    M = YEAR_IN_DAYS / 12
    Q = YEAR_IN_DAYS / 4
    Y = YEAR_IN_DAYS

    @classmethod
    def from_string(cls, n: str) -> "IntervalUnit":
        return cls[n.upper()]

    def timedelta(self) -> timedelta:
        return timedelta(days=self.value)

    @override
    def __str__(self) -> str:
        return self.name

    @override
    def __repr__(self) -> str:
        return f"IntervalUnit.{str(self)}"


@final
class Interval:
    """Mapping years, month, and quarter to real numbers of approximate days. Weeks and days mapped to integer. "

    >>> Interval(1, IntervalUnit.D).timedelta()
    datetime.timedelta(days=1)
    >>> Interval(1, IntervalUnit.W).timedelta()
    datetime.timedelta(days=7)
    >>> Interval(1, IntervalUnit.M).timedelta()
    datetime.timedelta(days=30, seconds=37843, microseconds=200000)
    >>> Interval(1, IntervalUnit.Q).timedelta()
    datetime.timedelta(days=91, seconds=27129, microseconds=600000)
    >>> Interval(1, IntervalUnit.Y).timedelta()
    datetime.timedelta(days=365, seconds=22118, microseconds=400000)
    """

    _P = "".join(p.name for p in IntervalUnit)
    FREQ_RE = re.compile(r"(\d+)([" + _P + _P.lower() + "])")

    multiplier: int
    period: IntervalUnit

    def __init__(self, multiplier: int, period: IntervalUnit) -> None:
        self.multiplier = multiplier
        self.period = period

    @classmethod
    def from_string_safe(cls, s: "Interval|str|None") -> "Interval | None":
        if s is None:
            return None
        if isinstance(s, Interval):
            return s
        return cls.from_string(s)

    @classmethod
    def from_string(cls, s: str) -> "Interval":
        m = cls.matcher(s)
        if m:
            n, p = m.groups()
            return cls(int(n), IntervalUnit.from_string(p))
        else:
            raise ValueError("Invalid frequency string", s)

    @classmethod
    def matcher(cls, s: str) -> re.Match[str] | None:
        return re.match(cls.FREQ_RE, s)

    def timedelta(self) -> timedelta:
        return self.multiplier * self.period.timedelta()

    def match(self, d: date, as_of: date) -> bool:
        return d <= as_of and d + self.timedelta() > as_of

    def find_file(
        self,
        path: Path,
        as_of: date | datetime,
        suffix: str = ".csv",
    ) -> Path | None:
        ff = list(
            reversed(
                sorted(
                    (date_from_name(f.name), f)
                    for f in path.glob(f"*{suffix}")
                    if re.match(r"^\d{8}", f.name[:8])
                )
            )
        )
        for d, f in ff:
            if d <= as_of:
                if self.match(d, as_of):
                    return f
                else:
                    break
        return None

    @override
    def __str__(self) -> str:
        return f"{self.multiplier}{self.period}"

    @override
    def __repr__(self) -> str:
        return f"Interval({self.multiplier}, {self.period!r})"


IntervalSafe = Annotated[
    Interval | str | None,
    BeforeValidator(Interval.from_string_safe),
    PlainSerializer(str_or_none, return_type=str),
    WithJsonSchema({"anyOf": [{"type": "string"}, {"type": "null"}]}),
]

KNOWN_TYPES.register(
    KnownTypeArgs(concrete_type=Interval, map_to_string=str),
)


def date_from_name(s: str) -> date:
    return date(int(s[:4]), int(s[4:6]), int(s[6:8]))


class Moment:
    """
    >>> m = Moment.start()
    >>> m = m.capture("instant")
    >>> tt.sleep(1)
    >>> m = m.capture("a second")
    >>> s = m.chain()
    >>> s.startswith('[start] 0.0'), 's-> [instant] 1.' in s , s.endswith('s-> [a second]')
    (True, True, True)
    """

    time: float
    name: str
    prev: "Moment | None"

    def __init__(self, name: str, prev: "Moment | None" = None) -> None:
        self.time = tt.time()
        self.name = name
        self.prev = prev

    @staticmethod
    def start():
        """capture the starting moment"""
        return Moment("start")

    def capture(self, name: str):
        """capture the named moment relative to this one"""
        return Moment(name, self)

    def elapsed(self):
        """return time in seconds since previous moment"""
        if self.prev is None:
            return 0
        return self.time - self.prev.time

    @override
    def __str__(self):
        return (
            f" {self.elapsed():.3f}s-> [{self.name}]" if self.prev is not None else f"[{self.name}]"
        )

    def chain(self) -> str:
        return str(self) if self.prev is None else self.prev.chain() + str(self)


class PeriodicTask:
    """
    A task that runs at a specified frequency in seconds.

    Use with `run_all()` to execute multiple tasks in an async event loop.
    """

    freq: int
    logic: Callable[[], Any]
    last_run: float | None = None

    def __init__(self, freq: int, logic: Callable[[], Any]) -> None:
        self.freq = freq
        self.logic = logic

    def is_due(self):
        return self.last_run is None or stime.time() - self.last_run > self.freq


def gcd_pair(a: int, b: int) -> int:
    """
    >>> gcd_pair(4, 6)
    2
    >>> gcd_pair(6*15, 6*7)
    6
    >>> gcd_pair(6,35)
    1
    """
    return abs(a) if b == 0 else gcd_pair(b, a % b)


def gcd(*nn: int) -> int:
    """
    >>> gcd(4)
    4
    >>> gcd(4, 6)
    2
    >>> gcd(6*15, 6*7)
    6
    >>> gcd(6,35)
    1
    >>> gcd(6*15, 6*7, 6*5)
    6
    >>> gcd(6*15, 6*7, 10)
    2
    >>> gcd(6*15, 6*7, 35)
    1
    >>> gcd()
    Traceback (most recent call last):
    ...
    IndexError: tuple index out of range
    """
    r = nn[0]
    for i in range(1, len(nn)):
        r = gcd_pair(r, nn[i])
    return r


def _collect_nothing(n: str, x: Any):  # pyright: ignore [reportUnusedParameter]
    pass  # pragma: no cover


async def run_all(
    *tasks: PeriodicTask,
    shutdown_event: asyncio.Event | None = None,
    collect_results: Callable[[str, Any], None] = _collect_nothing,
):
    """
    Run multiple periodic tasks in an async loop.

    Tasks are scheduled based on their `freq` (in seconds). The loop tick interval
    is the GCD of all task frequencies for efficiency. Set `shutdown_event` to
    signal graceful termination.
    """
    if shutdown_event is None:
        shutdown_event = asyncio.Event()
    if len(tasks) == 0:
        log.warning("No tasks to run")
        return
    tick = gcd(*[t.freq for t in tasks])
    loop = asyncio.get_running_loop()

    while True:
        start = stime.time()
        for t in tasks:
            if t.is_due():
                t.last_run = start

                try:
                    if inspect.iscoroutinefunction(t.logic):
                        r = await t.logic()
                    else:
                        r = await loop.run_in_executor(None, t.logic)
                except (Exception, asyncio.CancelledError) as _:
                    r = sys.exc_info()
                collect_results(t.logic.__name__, r)
            if shutdown_event.is_set():
                return
        elapsed = stime.time() - start
        await asyncio.sleep(tick - elapsed if elapsed < tick else 0)


def adjust_as_of_date(as_of_date: date | None) -> date:
    """
    >>> adjust_as_of_date(None) == date.today()
    True
    >>> adjust_as_of_date(date(2021, 1, 1)) == date(2021, 1, 1)
    True
    """
    return date.today() if as_of_date is None else as_of_date



import sqlite3
from datetime import date, timedelta
from typing import Literal

from pydantic import BaseModel, Field

from lythonic.config import Configurable
from lythonic.examples.cashflow_tracking import (
    SCHEMA,
    Account,
    CashAccountProjection,
    CashEvent,
    CashEventNotification,
    FlowItem,
    FlowProjection,
    Organization,
    ScheduledEvent,
)
from lythonic.state import DbConfig, DbFile
from lythonic.state.user import UserContext

PROJECTION_RANGE = timedelta(days=62)  # 2 months


class ApiContext(UserContext):
    as_of: date = Field(description="Date of the projection typically today")


EventAction = Literal["create_event", "imagine_event", "create_notification", "do_nothing"]


class CashAccountProjectionCalculator(BaseModel):
    cash_account: Account = Field(description="Cash account to calculate the projection for")
    cash_projection: CashAccountProjection | None = Field(
        default=None, description="Current projection for the cash account"
    )
    new_events: list[CashEvent] = Field(
        default_factory=list,
        description="List of events materialized and projected for the cash account",
    )
    new_notifications: list[CashEventNotification] = Field(
        default_factory=list,
        description="List of notifications for the events for the cash account",
    )

    def calculate_projection(
        self, conn: sqlite3.Connection, as_of: date, force: bool = False
    ) -> tuple[bool, CashAccountProjection]:
        need_recalculate = (
            force or self.cash_projection is None or self.cash_projection.as_of < as_of
        )
        if need_recalculate:
            # Look for the latest 'set_balance' CashEvent, and keep only the events after it to be considered for projection
            starting_from_date = as_of - PROJECTION_RANGE
            cash_events: list[CashEvent] = sorted(
                CashEvent.select(
                    conn, cash_acc_id=self.cash_account.acc_id, event_date__gte=starting_from_date
                ),
                key=lambda x: x.event_date,
            )
            set_balance_events: list[int] = [
                i for i, e in enumerate(cash_events) if e.event_type == "set_balance"
            ]
            assert set_balance_events, (
                f"No 'set_balance' CashEvent found starting from the date {starting_from_date}"
            )

            relevant_cash_events = cash_events[set_balance_events[-1] :]
            start_projection_date = cash_events[0].event_date
            end_projection_date = as_of + PROJECTION_RANGE

            relevant_notifications = CashEventNotification.select(
                conn,
                cash_acc_id=self.cash_account.acc_id,
                notification_date__gte=start_projection_date,
            )

            class EventProjection:
                event_date: date
                schedule: ScheduledEvent
                account: Account
                matched_cash_event: CashEvent | None = None
                matched_notification: CashEventNotification | None = None
                action: EventAction = "do_nothing"
                new_cash_event: CashEvent | None = None
                new_notification: CashEventNotification | None = None

                def __init__(self, event_date: date, schedule: ScheduledEvent, account: Account):
                    self.event_date = event_date
                    self.schedule = schedule
                    self.account = account
                    self.matched_cash_event = next(
                        (
                            e
                            for e in relevant_cash_events
                            if e.event_date == event_date and e.sch_id == schedule.sch_id
                        ),
                        None,
                    )
                    self.matched_notification = next(
                        (
                            n
                            for n in relevant_notifications
                            if n.note_date == event_date and n.sch_id == schedule.sch_id
                        ),
                        None,
                    )
                    if self.schedule.amount_type == "fixed" and self.matched_cash_event is None:
                        if self.event_date <= as_of:
                            self.action = "create_event"
                        else:
                            self.action = "imagine_event"
                        self.new_cash_event = CashEvent.from_scheduled_event(
                            self.schedule, self.event_date, self.account
                        )
                    elif (
                        self.schedule.amount_type == "variable"
                        and self.matched_cash_event is None
                        and self.matched_notification is None
                        and self.event_date
                        <= as_of + timedelta(days=self.schedule.reminder_days or 1)
                    ):
                        self.action = "create_notification"
                        self.new_notification = CashEventNotification.from_scheduled_event(
                            self.schedule, self.event_date
                        )
                    else:
                        self.action = "do_nothing"

                def flow_item(self) -> FlowItem | None:
                    if self.matched_cash_event is not None:
                        return FlowItem.from_cash_event(self.matched_cash_event)
                    elif self.new_cash_event is not None:
                        return FlowItem.from_cash_event(self.new_cash_event)
                    else:
                        return None

            scheduled_events = ScheduledEvent.select(
                conn, is_active=True, account_id=self.cash_account.acc_id
            )

            event_projections: list[EventProjection] = []

            accounts = {
                a.acc_id: a
                for a in Account.select(
                    conn, acc_id=[scheduled_event.acc_id for scheduled_event in scheduled_events]
                )
            }

            for scheduled_event in scheduled_events:
                account = accounts[scheduled_event.acc_id]
                for event_date in scheduled_event.get_event_dates_for_range(
                    start_projection_date, end_projection_date
                ):
                    event_projections.append(EventProjection(event_date, scheduled_event, account))

            left_over_events = {e.cash_id: e for e in relevant_cash_events}
            for event_projection in event_projections:
                if event_projection.matched_cash_event is not None:
                    left_over_events.pop(event_projection.matched_cash_event.cash_id)

            flow_items: list[FlowItem] = []
            for event_projection in event_projections:
                if (
                    event_projection.action == "create_event"
                    and event_projection.new_cash_event is not None
                ):
                    event_projection.new_cash_event.save(conn)
                elif (
                    event_projection.action == "create_notification"
                    and event_projection.new_notification is not None
                ):
                    event_projection.new_notification.save(conn)

                flow_item = event_projection.flow_item()
                if flow_item is not None:
                    flow_items.append(flow_item)
            flow_items.extend(FlowItem.from_cash_event(e) for e in left_over_events.values())
            flow_items.sort(key=lambda x: (x.event_date, x.event_type != "set_balance", -x.amount))
            assert flow_items[0].event_type == "set_balance", (
                "First flow item must be the set_balance event"
            )
            balance = flow_items[0].amount
            alert = False
            for flow_item in flow_items:
                flow_item.balance = balance
                if flow_item.balance < 0:
                    alert = True
                balance += flow_item.amount

            projection = FlowProjection(as_of=as_of, events=flow_items, alert=alert)

            self.cash_projection = CashAccountProjection(
                cash_acc_id=self.cash_account.acc_id,
                user_id=self.cash_account.user_id,
                projection=projection,
                as_of=as_of,
            )
            self.cash_projection.save(conn)
        assert self.cash_projection is not None

        return need_recalculate, self.cash_projection


class CashflowTrackingApi(Configurable[DbConfig]):
    db_file: DbFile

    def __init__(self, config: DbConfig) -> None:
        super().__init__(config)
        self.db_file = DbFile(config.db_path, SCHEMA)
        self.db_file.check(ensure=True)

    def save_organization(self, org: Organization) -> Organization:
        with self.db_file.open() as conn:
            org.save(conn)
            return org

    def list_organizations(self, is_hidden: bool = False) -> list[Organization]:
        with self.db_file.open() as conn:
            return Organization.select(conn, is_hidden=is_hidden)

    def get_organization(self, org_id: int) -> Organization:
        with self.db_file.open() as conn:
            org = Organization.load_by_id(conn, org_id)
            if org is None:
                raise ValueError(f"Organization with id {org_id} not found in database")
            return org

    def get_all_accounts(
        self,
        ctx: ApiContext,
        org_id: int | None = None,
    ) -> list[Account]:
        with self.db_file.open() as conn:
            if org_id is not None:
                return Account.select(conn, user_id=ctx.user.user_id, org_id=org_id)
            else:
                return Account.select(conn, user_id=ctx.user.user_id)

    def save_account(self, ctx: ApiContext, account: Account) -> Account:
        account.user_id = ctx.user.user_id
        with self.db_file.open() as conn:
            if account.acc_id != -1:
                assert (
                    len(account.select(conn, acc_id=account.acc_id, user_id=ctx.user.user_id)) == 1
                ), "Account not found"
            account.save(conn)
            return account

    def delete_account(self, ctx: ApiContext, account: Account) -> None:
        account.user_id = ctx.user.user_id
        assert account.acc_id != -1, "Account ID is -1"
        with self.db_file.open() as conn:
            account.end_date = ctx.as_of
            assert (
                len(account.select(conn, acc_id=account.acc_id, user_id=ctx.user.user_id)) == 1
            ), "Account not found"
            account.save(conn)

    def get_scheduled_events(
        self, ctx: ApiContext, org_id: int
    ) -> list[tuple[ScheduledEvent, Account]]:
        with self.db_file.open() as conn:
            active_accounts = {
                a.acc_id: a
                for a in Account.filter_active(
                    Account.select(conn, user_id=ctx.user.user_id, org_id=org_id), ctx.as_of
                )
            }
            return [
                (se, active_accounts[se.acc_id])
                for se in ScheduledEvent.select(
                    conn,
                    is_active=True,
                    user_id=ctx.user.user_id,
                    account_id=list(active_accounts.keys()),
                )
            ]

    def calculate_cash_account_projections(
        self, ctx: ApiContext, org_id: int, force: bool = False
    ) -> tuple[bool, dict[int, CashAccountProjection]]:
        with self.db_file.open() as conn:
            calculators: dict[int, CashAccountProjectionCalculator] = {
                ca.acc_id: CashAccountProjectionCalculator(cash_account=ca)
                for ca in Account.filter_active(
                    Account.select(
                        conn, account_type="cash", user_id=ctx.user.user_id, org_id=org_id
                    ),
                    ctx.as_of,
                )
            }
            for cap in CashAccountProjection.select(conn, cash_acc_id=list(calculators.keys())):
                calculators[cap.cash_acc_id].cash_projection = cap
            results: list[tuple[bool, CashAccountProjection]] = [
                calculator.calculate_projection(conn, ctx.as_of, force)
                for calculator in calculators.values()
            ]
            return any(map(lambda x: x[0], results)), {r[1].cash_acc_id: r[1] for r in results}



"""
User: Multi-tenant user ownership patterns for database models.

This module provides base classes for building multi-tenant applications where
records are owned by users and access is controlled accordingly.

## Core Classes

- `User`: Base user model with `user_id`, `info`, and `created_at`
- `UserContext`: Context object passed to operations requiring user scope
- `UserOwned`: Base class for models that belong to a user

## Usage

```python
from lythonic.state.user import User, UserContext, UserOwned, UserInfo
from pydantic import Field

class MyUserInfo(UserInfo):
    name: str
    email: str

class Document(UserOwned["Document"]):
    doc_id: int = Field(default=-1, description="(PK)")
    title: str

# Create user context
user = User(user_id=1, info=MyUserInfo(name="Alice", email="alice@example.com"))
ctx = UserContext(user=user)

# Save with user context (enforces ownership)
doc = Document(title="My Doc")
doc.save_with_ctx(ctx, conn)

# Load with user context (only returns user's records)
doc = Document.load_by_id_with_ctx(conn, ctx, doc_id=1)
```

The `UserOwned` class overrides `save()` and `load_by_id()` to require
`UserContext`, ensuring all database operations are scoped to the current user.
"""

from __future__ import annotations

import sqlite3
from datetime import datetime
from typing import Any, TypeVar

from pydantic import BaseModel, Field
from typing_extensions import override

from lythonic import GlobalRef, utc_now
from lythonic.state import DbModel
from lythonic.types import JsonBase


class UserInfo(JsonBase):
    pass


class User(DbModel["User"]):
    user_id: int = Field(default=-1, description="(PK) Unique identifier for the user")
    info: UserInfo = Field(description="User information object")
    created_at: datetime = Field(
        default_factory=utc_now, description="Date and time when the user was created"
    )


class UserContext(BaseModel):
    user: User


UO = TypeVar("UO", bound="UserOwned")  # pyright: ignore [reportMissingTypeArgument]


class UserOwned(DbModel[UO]):
    user_id: int = Field(default=-1, description="(FK:User.user_id) Reference to the user")

    @override
    def save(self, conn: sqlite3.Connection) -> None:
        raise NotImplementedError("Use save_with_ctx instead")

    @override
    @classmethod
    def load_by_id(cls: type[UO], conn: sqlite3.Connection, id: int) -> UO | None:
        raise NotImplementedError("Use load_by_id_with_ctx instead")

    def save_with_ctx(self, ctx: UserContext, conn: sqlite3.Connection):
        self.user_id = ctx.user.user_id
        cls = self.__class__
        pks = cls._choose_fields(lambda fi: fi.primary_key)
        assert len(pks) == 1
        pk = pks[0]
        pk_val = getattr(self, pk.name)
        if pk_val == -1:
            self.insert(conn, auto_increment=True)
            return
        n_updated = self.update(conn, user_ctx=ctx, **{pk.name: pk_val})
        if n_updated == 0:
            raise ValueError(
                "Possible Access violation:",
                " Record does not exist or belong other user then the one in context:",
                ctx.user.user_id,
            )
        else:
            assert n_updated == 1

    @override
    @classmethod
    def _prepare_where(cls, conn: sqlite3.Connection, **filters: Any) -> DbModel._WhereBased:
        assert "user_ctx" in filters, f"user_ctx:{GlobalRef(UserContext)!s} is required"
        user_ctx = filters.pop("user_ctx")
        assert isinstance(user_ctx, UserContext), (
            f"user_ctx:{GlobalRef(UserContext)!s} must be a UserContext"
        )
        filters["user_id"] = user_ctx.user.user_id
        return super()._prepare_where(conn, **filters)

    @classmethod
    def load_by_id_with_ctx(
        cls: type[UO], conn: sqlite3.Connection, user_ctx: UserContext, id: int
    ) -> UO | None:
        rr: list[UO] = cls.select(conn, user_ctx=user_ctx, **{cls._ensure_pk().name: id})
        assert len(rr) <= 1
        return rr[0] if rr else None



import logging
from collections.abc import Generator
from datetime import UTC, datetime
from pathlib import Path

import pytest

import tests.rag_schema as rag
from lythonic.examples.cashflow_tracking import SCHEMA as acc
from lythonic.misc import tabula_rasa_path
from lythonic.state import (
    FieldInfo,
    open_sqlite_db,
)
from lythonic.types import KnownType


def test_type_info():
    assert FieldInfo.build("timestamp", rag.RagAction.model_fields["timestamp"]) == (
        "timestamp",
        KnownType.ensure("datetime"),
        "When the attempt was made",
        False,
        False,
        None,
        None,
    )


rag_ddls = (
    "CREATE TABLE RagSource (source_id INTEGER PRIMARY KEY, absolute_path TEXT NOT NULL)",
    "CREATE TABLE RagAction (action_id INTEGER PRIMARY KEY, source_id INTEGER NOT NULL REFERENCES RagSource(source_id), timestamp TEXT NOT NULL, n_chunks INTEGER NOT NULL, error TEXT, sha256 TEXT NOT NULL)",
    "CREATE TABLE RagActionCollection (action_id INTEGER NOT NULL REFERENCES RagAction(action_id), action TEXT NOT NULL CHECK (action IN ('new', 'update', 'delete')), collection TEXT NOT NULL, timestamp TEXT NOT NULL)",
    "CREATE TABLE ConvoMessage (role TEXT NOT NULL CHECK (role IN ('system', 'user', 'assistant', 'tool')), tool_call_id TEXT, content TEXT NOT NULL, finish_reason TEXT NOT NULL CHECK (finish_reason IN ('stop', 'length', 'content_filter', 'null')), message_id INTEGER PRIMARY KEY, session_id INTEGER NOT NULL REFERENCES ConvoSession(session_id), captured TEXT NOT NULL)",
    "CREATE TABLE ConvoSession (session_id INTEGER PRIMARY KEY, created TEXT NOT NULL, updated TEXT NOT NULL, model TEXT NOT NULL, user_id TEXT, session_type TEXT NOT NULL CHECK (session_type IN ('active', 'completed', 'failed', 'archived')))",
)

acc_ddls = (
    "CREATE TABLE User (user_id INTEGER PRIMARY KEY, info TEXT NOT NULL, created_at TEXT NOT NULL)",
    "CREATE TABLE Organization (user_id INTEGER NOT NULL REFERENCES User(user_id), org_id INTEGER PRIMARY KEY, name TEXT NOT NULL, created_at TEXT NOT NULL, is_hidden INTEGER NOT NULL)",
    "CREATE TABLE Account (user_id INTEGER NOT NULL REFERENCES User(user_id), acc_id INTEGER PRIMARY KEY, org_id INTEGER NOT NULL REFERENCES Organization(org_id), name TEXT NOT NULL, account_type TEXT NOT NULL CHECK (account_type IN ('cash', 'credit_card')), account_kind TEXT, start_date TEXT NOT NULL, end_date TEXT, linked_cash_acc_id INTEGER REFERENCES Account(acc_id), created_at TEXT NOT NULL)",
    "CREATE TABLE ScheduledEvent (user_id INTEGER NOT NULL REFERENCES User(user_id), sch_id INTEGER PRIMARY KEY, acc_id INTEGER NOT NULL REFERENCES Account(acc_id), description_template TEXT NOT NULL, event_type TEXT NOT NULL CHECK (event_type IN ('deposit', 'payment', 'set_balance')), amount_type TEXT NOT NULL CHECK (amount_type IN ('fixed', 'variable')), amount REAL, frequency TEXT NOT NULL CHECK (frequency IN ('weekly', 'monthly', 'quarterly', 'annually')), day_in_the_cycle INTEGER NOT NULL, reminder_days INTEGER, root_version_id INTEGER REFERENCES ScheduledEvent(sch_id), is_active INTEGER NOT NULL, start_date TEXT NOT NULL, end_date TEXT, modified_at TEXT NOT NULL)",
    "CREATE TABLE CashEvent (user_id INTEGER NOT NULL REFERENCES User(user_id), cash_id INTEGER PRIMARY KEY, cash_acc_id INTEGER NOT NULL REFERENCES Account(acc_id), related_acc_id INTEGER REFERENCES Account(acc_id), sch_id INTEGER REFERENCES ScheduledEvent(sch_id), event_type TEXT NOT NULL CHECK (event_type IN ('deposit', 'payment', 'set_balance')), event_date TEXT NOT NULL, amount REAL NOT NULL, description TEXT, modified_at TEXT NOT NULL)",
    "CREATE TABLE CashEventNotification (user_id INTEGER NOT NULL REFERENCES User(user_id), note_id INTEGER PRIMARY KEY, sch_id INTEGER NOT NULL REFERENCES ScheduledEvent(sch_id), note_date TEXT NOT NULL, muted INTEGER NOT NULL, event_date TEXT NOT NULL)",
    "CREATE TABLE CashAccountProjection (user_id INTEGER NOT NULL REFERENCES User(user_id), cash_acc_id INTEGER PRIMARY KEY REFERENCES Account(acc_id), projection TEXT, as_of TEXT NOT NULL, modified_at TEXT NOT NULL)",
)


def test_ddl():
    assert rag.RagSource.create_ddl() == rag_ddls[0]
    assert rag.RagAction.create_ddl() == rag_ddls[1]
    assert rag.RagActionCollection.create_ddl() == rag_ddls[2]
    assert rag.ConvoMessage.create_ddl() == rag_ddls[3]
    assert rag.ConvoSession.create_ddl() == rag_ddls[4]

    assert acc_ddls == tuple(t.create_ddl() for t in acc.tables)


# test db cleanup
rag_db_path = tabula_rasa_path(Path("build/tests/rag.db"))
acc_db_path = tabula_rasa_path(Path("build/tests/acc.db"))


def search_caplog(
    caplog: pytest.LogCaptureFixture,
    prefix: str,
    suffix: str | None = None,
    category: str | None = None,
) -> Generator[str, None, None]:
    for r in caplog.records:
        m = r.getMessage()
        if category is None or r.name == category:
            if m.startswith(prefix):
                s = m[len(prefix) :]
                if suffix is not None:
                    idx = s.index(suffix)
                    if idx != -1:
                        s = s[:idx]
                yield s


def test_rag_db(caplog: pytest.LogCaptureFixture):
    caplog.set_level(logging.DEBUG)
    rag.SCHEMA.create_schema(rag_db_path)
    extract = tuple(search_caplog(caplog, "execute: ", category="lythonic.state"))
    print(extract)
    assert extract == rag_ddls


def test_acc_db(caplog: pytest.LogCaptureFixture):
    caplog.set_level(logging.DEBUG)
    acc.create_schema(acc_db_path)
    extract = tuple(search_caplog(caplog, "execute: ", category="lythonic.state"))
    print(extract)
    assert extract == acc_ddls


def test_add_attempt():
    with open_sqlite_db(rag_db_path) as conn:
        select = rag.select_all_active_sources(conn)
        assert len(select) == 0
        s = rag.RagSource(absolute_path=Path("test.txt"))
        assert s.source_id == -1
        s.save(conn)
        assert s.source_id != -1
        s_loaded = rag.RagSource.load_by_id(conn, s.source_id)
        assert s_loaded is not None
        assert s_loaded == s
        assert isinstance(s_loaded.absolute_path, Path)
        a1 = rag.RagAction(
            source_id=s.source_id,
            n_chunks=1,
            error=None,
            sha256="1234567890",
        )
        assert a1.action_id == -1
        a1.save(conn)

        assert a1.action_id != -1
        c1 = rag.RagActionCollection(
            action_id=a1.action_id,
            action="new",
            collection="test",
        )
        c1.insert(conn)

        a2 = rag.RagAction(
            source_id=s.source_id,
            n_chunks=1,
            error=None,
            sha256="6789012345",
        )
        assert a2.action_id == -1
        a2.save(conn)

        assert a2.action_id != -1
        assert a2.action_id != a1.action_id
        c2 = rag.RagActionCollection(
            action_id=a2.action_id,
            action="new",
            collection="test",
        )
        c2.insert(conn)

        a2.timestamp = datetime.now(tz=UTC)
        a2.sha256 = "6172839405"
        a2.save(conn)

        select = rag.select_all_active_sources(conn)
        if len(select) != 1:
            print(f"{select=}")
            raise AssertionError(f"{select=}")

        assert len(select[0][2]) == 1

        a3 = rag.RagAction(
            action_id=6,
            source_id=s.source_id,
            n_chunks=6,
            error=None,
            sha256="1122334455",
        )
        a3.save(conn)

        c3 = rag.RagActionCollection(
            action_id=a3.action_id,
            action="update",
            collection="test",
        )
        c3.insert(conn)

        c4 = rag.RagActionCollection(
            action_id=a3.action_id,
            action="new",
            collection="test2",
        )
        c4.insert(conn)

        a1_loaded = rag.RagAction.load_by_id(conn, a1.action_id)
        a2_loaded = rag.RagAction.load_by_id(conn, a2.action_id)
        a3_loaded = rag.RagAction.load_by_id(conn, a3.action_id)
        nothing_loaded = rag.RagAction.load_by_id(conn, 7)
        assert nothing_loaded is None
        assert a1_loaded is not None
        assert a2_loaded is not None
        assert a3_loaded is not None
        assert a1_loaded == a1, f"{a1_loaded=} != {a1=}"
        assert a2_loaded == a2, f"{a2_loaded=} != {a2=}"
        assert a3_loaded == a3, f"{a3_loaded=} != {a3=}"
        assert isinstance(a1_loaded.timestamp, datetime)
        assert isinstance(a2_loaded.timestamp, datetime)
        assert isinstance(a3_loaded.timestamp, datetime)
        select = rag.RagAction.select(conn, source_id=s.source_id)
        assert len(select) == 3
        assert a1 in select
        assert a2 in select
        assert a3 in select
        select = rag.select_all_active_sources(conn)
        assert len(select) == 1
        s_loaded2, a_loaded2, c_loaded2 = select[0]
        assert s_loaded2 == s
        assert a_loaded2 == a3
        assert isinstance(c_loaded2, list)
        assert len(c_loaded2) == 2
        conn.commit()



import logging
from datetime import date
from pathlib import Path

import pytest

from lythonic.examples.cashflow_tracking import SCHEMA as CFT_SCHEMA
from lythonic.examples.cashflow_tracking import Account, Organization, ScheduledEvent
from lythonic.misc import tabula_rasa_path
from lythonic.state import open_sqlite_db
from lythonic.state.user import User, UserContext, UserInfo
from lythonic.types import base_model_to_json, json_dumps
from tests.test_sqlite import acc_ddls, search_caplog

acc_db_path = tabula_rasa_path(Path("build/tests/cft_api.db"))


def test_acc_db(caplog: pytest.LogCaptureFixture):
    caplog.set_level(logging.DEBUG)
    CFT_SCHEMA.create_schema(acc_db_path)
    extract = tuple(search_caplog(caplog, "execute: ", category="lythonic.state"))
    print(extract)
    assert extract == acc_ddls


def test_create_user(caplog: pytest.LogCaptureFixture):
    caplog.set_level(logging.DEBUG)
    with open_sqlite_db(acc_db_path) as conn:
        u = User(info=UserInfo())
        u.save(conn)
        loaded = User.load_by_id(conn, u.user_id)
        assert loaded == u
        conn.commit()
    extract = tuple(search_caplog(caplog, "execute: ", category="lythonic.state"))
    assert extract == (
        f"INSERT INTO User (info, created_at) VALUES (?, ?) -- with args: {([json_dumps(base_model_to_json(u.info)), u.created_at.isoformat()],)}",
        f"SELECT user_id, info, created_at FROM User WHERE user_id = ? -- with args: {([u.user_id],)}",
    )


def test_user_owned_objects(caplog: pytest.LogCaptureFixture):
    caplog.set_level(logging.DEBUG)
    with open_sqlite_db(acc_db_path) as conn:
        u2 = User(info=UserInfo())
        u2.save(conn)
        users: dict[int, User] = {u.user_id: u for u in User.select(conn)}
        assert len(users) == 2
        assert users.keys() == {1, 2}
        ctxs: dict[int, UserContext] = {u.user_id: UserContext(user=u) for u in users.values()}

        o = Organization(name="Personal")
        o.save_with_ctx(ctxs[2], conn)
        loaded = Organization.load_by_id_with_ctx(conn, ctxs[2], o.org_id)
        assert loaded == o

        bank1 = Account(
            name="BankOne Checking",
            account_type="cash",
            org_id=o.org_id,
            start_date=date(2025, 1, 1),
        )
        bank1.save_with_ctx(ctxs[2], conn)
        loaded = Account.load_by_id_with_ctx(conn, ctxs[2], bank1.acc_id)
        assert loaded == bank1

        sch1 = ScheduledEvent(
            acc_id=bank1.acc_id,
            description_template="Set balalnce at the beginning of the month",
            event_type="set_balance",
            amount_type="variable",
            amount=None,
            frequency="monthly",
            day_in_the_cycle=0,
            reminder_days=2,
        )
        with pytest.raises(NotImplementedError):
            sch1.save(conn)
        with pytest.raises(NotImplementedError):
            sch1.load_by_id(conn, sch1.sch_id)
        with pytest.raises(AssertionError):
            sch1.select(conn, sch_id=sch1.sch_id)
        sch1.save_versioned(ctxs[2], conn, date(2025, 1, 1))
        sch1.reminder_days = 3
        sch1.save_versioned(ctxs[2], conn, date(2025, 1, 2))
        sch1.reminder_days = 4
        sch1.save_versioned(ctxs[2], conn, date(2025, 1, 2))
        rr = ScheduledEvent.select(conn, user_ctx=ctxs[2], acc_id=bank1.acc_id)
        assert len(rr) == 2
        assert {sch.is_active: sch.reminder_days for sch in rr} == {True: 4, False: 2}

        sch2 = ScheduledEvent(
            acc_id=bank1.acc_id,
            description_template="Mid month salary deposit",
            event_type="deposit",
            amount_type="fixed",
            amount=2000,
            frequency="monthly",
            day_in_the_cycle=14,
        )
        sch2.save_versioned(ctxs[2], conn, date(2025, 1, 1))
        sch2.day_in_the_cycle = 15
        with pytest.raises(ValueError) as e:
            sch2.save_versioned(ctxs[1], conn, date(2025, 1, 2))
        assert str(e.value.args[0]) == "Cannot save prev version:"

        with pytest.raises(ValueError) as e:
            sch2.enddate_record(ctxs[1], conn, date(2025, 1, 2))
        assert str(e.value.args[0]) == "Possible Access violation:"

        rr = list(sch2.get_event_dates_for_range(date(2025, 1, 1), date(2025, 3, 31)))

        assert rr == [date(2025, 1, 16), date(2025, 2, 16), date(2025, 3, 16)]

        rr = list(sch1.get_event_dates_for_range(date(2025, 1, 1), date(2025, 3, 31)))

        assert rr == [date(2025, 1, 1), date(2025, 2, 1), date(2025, 3, 1)]

        conn.commit()



"""
Types: Type mapping system for converting between Python, JSON, and SQLite types.

This module provides a registry of known types and their conversion functions
for seamless data transformation between different representations.

## Core Concepts

- `KnownType`: A registered type with bidirectional mappings (string, JSON, DB)
- `KnownTypesMap`: Registry of all known types, accessed via `KNOWN_TYPES`
- `JsonBase`: Pydantic BaseModel with type discrimination via `type_gref`

## Type Mappings

Each `KnownType` provides three `MapPair`s for converting to/from:
- **string**: For CLI arguments, environment variables, config files
- **json**: For JSON serialization (API responses, storage)
- **db**: For SQLite storage (maps to INTEGER, REAL, TEXT, or BLOB)

## Built-in Types

Primitives: `int`, `float`, `str`, `bool`, `bytes`
Dates: `date`, `datetime` (ISO format strings)
Enums: `Enum` (string values), `IntEnum` (integer values)
Pydantic: `BaseModel`, `JsonBase` (JSON dicts)
Paths: `Path` (string representation)

## Usage

```python
from lythonic.types import KNOWN_TYPES, KnownType

# Resolve a type
kt = KNOWN_TYPES.resolve_type(datetime)

# Convert from string
dt = kt.string.map_from("2024-01-15T10:30:00")

# Convert to JSON
json_val = kt.json.map_to(dt)  # "2024-01-15T10:30:00"

# Convert to DB
db_val = kt.db.map_to(dt)  # "2024-01-15T10:30:00"
```

## Registering Custom Types

```python
from lythonic.types import KNOWN_TYPES, KnownTypeArgs

KNOWN_TYPES.register(
    KnownTypeArgs(
        concrete_type=MyType,
        map_from_string=MyType.parse,
        map_to_string=str,
    )
)
```

## JsonBase

`JsonBase` extends Pydantic's `BaseModel` with automatic type discrimination:

```python
from lythonic.types import JsonBase

class Dog(JsonBase):
    name: str

class Cat(JsonBase):
    name: str

# Serialize includes type reference
dog = Dog(name="Rex")
data = dog.to_json()  # {"type_gref": "mymodule:Dog", "name": "Rex"}

# Deserialize resolves correct type
animal = JsonBase.from_json(data)  # Returns Dog instance
```
"""

from __future__ import annotations

import base64
import inspect
import json as _json
from collections.abc import Callable
from datetime import date, datetime
from enum import Enum, IntEnum
from pathlib import Path
from typing import Any, Generic, Self, TypeVar, cast

from pydantic import (
    BaseModel,
    model_validator,
)
from typing_extensions import override

from lythonic import GlobalRef, GRef, str_or_none

# using this to allow for monkeypatching for NumPy
json_loads = _json.loads
json_dumps = _json.dumps


def base_model_to_json(bm: BaseModel) -> dict[str, Any]:
    return bm.model_dump(mode="json")


class JsonBase(BaseModel):
    type_gref: GRef | None = None

    @model_validator(mode="before")
    @classmethod
    def set_gref(cls, data: dict[str, Any]) -> dict[str, Any]:
        data["type_gref"] = str(GlobalRef(cls))
        return data

    @classmethod
    def from_json(cls, json: dict[str, Any]) -> Self:
        if "type_gref" in json:
            ref = GlobalRef(json["type_gref"])
            cls_from_ref = ref.get_instance()
            if issubclass(cls_from_ref, cls):
                return cls_from_ref.model_validate(json)
        return cls.model_validate(json)

    def to_json(self) -> dict[str, Any]:
        return base_model_to_json(self)


def encode_base64(bb: bytes) -> str:
    return base64.b64encode(bb).decode()


def ensure_bytes(input: str | bytes) -> bytes:
    if isinstance(input, str):
        input = base64.b64decode(input)
    return input


def do_identity(x: Any) -> Any:
    return x


def passthru_none(
    *fns: Callable[[Any], Any], if_none: Callable[[Any], Any] | None = None
) -> Callable[[Any], Any]:
    if if_none is not None:
        fn = if_none
    else:
        if not fns:
            return do_identity
        if len(fns) == 1:
            fn = fns[0]
            if fn is str:
                return str_or_none
        else:

            def do_combined(x: Any) -> Any:
                r = x
                previous: Callable[[Any], Any] | None = None
                for f in fns:
                    if previous != f:
                        r = f(r)
                    previous = f
                return r

            fn = do_combined
    if (
        fn.__name__ == "do_someCaZieSoNo1come_up_with_this_name"
        or fn == do_identity
        or fn == str_or_none
    ):
        return fn

    def do_someCaZieSoNo1come_up_with_this_name(x: Any) -> Any:
        if x is None:
            return x
        else:
            return fn(x)

    return do_someCaZieSoNo1come_up_with_this_name


T = TypeVar("T")


class MapPair(Generic[T]):
    target_type: type
    map_from: Callable[[T], Any]
    map_to: Callable[[Any], T]

    def __init__(
        self, map_from: Callable[[T], Any], map_to: Callable[[Any], T], target_type: type
    ) -> None:
        self.map_from = map_from
        self.map_to = map_to
        self.target_type = target_type


def is_primitive(type_: type) -> bool:
    """could be used in json and db as it is"""
    return type_ in (int, float, bool, str)


class DbTypeInfo(Enum):
    def __init__(self, input_types: tuple[type, ...], target_type: type):
        assert isinstance(input_types, tuple), f"input_types must be tuple, not {type(input_types)}"
        cls = self.__class__
        if not hasattr(cls, "_value2member_map_"):
            cls._value2member_map_ = {}
        member_map = cls._value2member_map_
        for k in input_types:
            assert k not in member_map, f"Duplicate input type {k} in {self.name}"
            member_map[k] = self
        self.input_types = input_types
        self.target_type = target_type

    INTEGER = ((int, bool), int)
    REAL = ((float,), float)
    TEXT = ((str,), str)
    BLOB = ((bytes,), bytes)

    @classmethod
    def from_type(cls, t: type) -> DbTypeInfo:
        return cast(
            DbTypeInfo,
            cls._value2member_map_[t],
        )

    @classmethod
    def is_db_type(cls, type_: type) -> bool:
        """could be used in db as it is"""
        return type_ in cls._value2member_map_


class KnownTypeArgs(BaseModel):
    map_from_string: Callable[[str], Any] | None = None
    map_to_string: Callable[[Any], str] | None = None
    map_from_json: Callable[[Any], Any] | None = None
    map_to_json: Callable[[Any], Any] | None = None
    map_from_db: Callable[[Any], Any] | None = None
    map_to_db: Callable[[Any], Any] | None = None
    db_type: type | None = None
    json_type: type | None = None
    concrete_type: type | None = None
    abstract_type: type | None = None
    is_factory: bool = False
    name: str | None = None
    aliases: list[str] | None = None

    def get_type(self) -> type:
        assert self.abstract_type is not None or self.concrete_type is not None
        if self.concrete_type is not None:
            assert self.abstract_type is None and not self.is_factory, (
                "concrete types cannot be factories"
            )
            return self.concrete_type
        assert self.abstract_type is not None, " Redundant but no other way to shut up pyright"
        return self.abstract_type

    def is_abstract(self) -> bool:
        return self.abstract_type is self.get_type()

    def build_concrete_type(self, concrete_type: type) -> Self:
        assert self.is_factory, "only factory types can build concrete types"
        super_type = self.get_type()
        assert issubclass(concrete_type, super_type), (
            f"concrete type {concrete_type} is not a subclass of abstract type {super_type}"
        )
        clone = self.model_copy()
        clone.concrete_type = concrete_type
        clone.abstract_type = None
        clone.is_factory = False

        def remap_constructor_from_super_type_to_concrete_type(
            x: Callable[[Any], Any] | None,
        ) -> Callable[[Any], Any] | None:
            if x is None:
                return None
            if x is super_type:  # switch to constructor of concrete type
                return concrete_type
            if (
                inspect.ismethod(x) and x.__self__ is super_type
            ):  # switch to class  method of concrete type
                return getattr(concrete_type, x.__name__)
            return x  # leave as is

        clone.map_from_string = remap_constructor_from_super_type_to_concrete_type(
            clone.map_from_string
        )
        clone.map_from_json = remap_constructor_from_super_type_to_concrete_type(
            clone.map_from_json
        )
        clone.map_from_db = remap_constructor_from_super_type_to_concrete_type(clone.map_from_db)
        return clone

    def _set_string_defaults(self) -> None:
        if self.db_type is None:
            self.db_type = str
        if self.json_type is None:
            self.json_type = str
        self.map_from_string = passthru_none(self.get_type(), if_none=self.map_from_string)
        self.map_to_string = passthru_none(str, if_none=self.map_to_string)
        self.map_from_json = passthru_none(self.map_from_string, if_none=self.map_from_json)
        self.map_to_json = passthru_none(self.map_to_string, if_none=self.map_to_json)
        self.map_from_db = passthru_none(self.map_from_string, if_none=self.map_from_db)
        self.map_to_db = passthru_none(self.map_to_string, if_none=self.map_to_db)

    def _set_db_defaults(self) -> None:
        assert self.db_type is not None and self.map_from_db is not None
        self.map_to_db = passthru_none(self.db_type, if_none=self.map_to_db)
        if self.db_type is bytes:
            if self.json_type is None:
                self.json_type = str
            self.map_from_json = passthru_none(ensure_bytes, if_none=self.map_from_json)
            self.map_to_json = passthru_none(encode_base64, if_none=self.map_to_json)
            self.map_from_string = passthru_none(
                ensure_bytes, self.map_from_db, if_none=self.map_from_string
            )
            self.map_to_string = passthru_none(
                self.map_to_db, encode_base64, if_none=self.map_to_string
            )
        else:
            if self.json_type is None:
                self.json_type = self.db_type
            self.map_from_json = passthru_none(self.map_from_db, if_none=self.map_from_json)
            self.map_to_json = passthru_none(self.map_to_db, if_none=self.map_to_json)
            self.map_from_string = passthru_none(
                self.db_type, self.map_from_db, if_none=self.map_from_string
            )
            self.map_to_string = passthru_none(self.map_to_db, str, if_none=self.map_to_string)

    def _set_json_defaults(self) -> bool:
        if self.map_from_json is not None and self.map_to_json is not None:
            if self.db_type is None:
                self.db_type = str
            if self.json_type is None:
                self.json_type = dict
            self.map_from_json = passthru_none(self.map_from_json)
            self.map_to_json = passthru_none(self.map_to_json)
            self.map_from_string = passthru_none(
                json_loads, self.map_from_json, if_none=self.map_from_string
            )
            self.map_to_string = passthru_none(
                self.map_to_json, json_dumps, if_none=self.map_to_string
            )
            self.map_from_db = passthru_none(self.map_from_string, if_none=self.map_from_db)
            self.map_to_db = passthru_none(self.map_to_string, if_none=self.map_to_db)
            return True
        return False

    def resolve_the_rest(self) -> Self:
        assert not self.is_factory, "factory types cannot resolve the rest"
        if self.concrete_type is not None:
            assert self.abstract_type is None
            if self.map_from_string is not None or self.map_to_string is not None:
                self._set_string_defaults()
            else:
                if self.db_type is None and DbTypeInfo.is_db_type(self.concrete_type):  # pyright: ignore
                    self.db_type = self.concrete_type
                if self.db_type is not None:
                    self.map_from_db = passthru_none(
                        self.db_type, self.concrete_type, if_none=self.map_from_db
                    )
                    self._set_db_defaults()
                else:
                    self._set_json_defaults()

        else:
            assert self.abstract_type is not None
            assert (
                self.map_from_string is not None
                or self.map_from_json is not None
                or self.map_from_db is not None
            )
            if self.map_from_string is not None:
                self._set_string_defaults()
            elif self._set_json_defaults():
                pass
            elif self.db_type is not None:
                assert self.map_from_db is not None
                self.map_from_db = passthru_none(self.map_from_db)
                self._set_db_defaults()

        return self


class KnownType:
    name: str
    aliases: set[str]
    type_: type
    is_abstract: bool
    string: MapPair[str]
    json: MapPair[Any]
    db: MapPair[Any]

    @classmethod
    def ensure(cls, type_: type | str | KnownType) -> KnownType:
        if isinstance(type_, KnownType):
            return type_
        return KNOWN_TYPES.resolve_type(type_)

    def __init__(
        self,
        args: KnownTypeArgs,
    ) -> None:
        assert not args.is_factory, "factory types cannot be initialized directly"
        self.type_ = args.get_type()
        self.is_abstract = args.is_abstract()
        args.resolve_the_rest()
        assert (
            args.db_type is not None
            and args.map_from_db is not None
            and args.map_to_db is not None
            and args.json_type is not None
            and args.map_from_json is not None
            and args.map_to_json is not None
            and args.map_from_string is not None
            and args.map_to_string is not None
        ), f"{self.type_} has not been resolved"

        self.string = MapPair(args.map_from_string, args.map_to_string, str)  # pyright: ignore
        self.json = MapPair(args.map_from_json, args.map_to_json, args.json_type)
        self.db = MapPair(args.map_from_db, args.map_to_db, args.db_type)
        self.name = (args.name if args.name is not None else self.type_.__name__).lower()
        self.aliases = set()
        self.aliases.add(self.name)
        if args.aliases is not None:
            for alias in args.aliases:
                self.aliases.add(alias.lower())

    def get_type(self) -> type:
        return self.type_

    @property
    def db_type_info(self) -> DbTypeInfo:
        return DbTypeInfo.from_type(self.db.target_type)

    @override
    def __repr__(self) -> str:
        return f"KnownType({self.name!r}, {self.type_!r})"


class AbstractTypeHeap:
    type_: type
    ktype: KnownType | KnownTypeArgs | None
    parent: AbstractTypeHeap | None
    children: list[AbstractTypeHeap]

    def __init__(
        self,
        type_: type,
        ktype: KnownType | KnownTypeArgs | None = None,
        parent: AbstractTypeHeap | None = None,
    ) -> None:
        self.type_ = type_
        self.ktype = ktype
        self.parent = parent
        self.children = []
        if parent is not None:
            still_parents_children: list[AbstractTypeHeap] = []
            for child in parent.children:
                m = self.matches_exactly(child.type_)
                assert m is not True
                if m is None:
                    still_parents_children.append(child)
                else:
                    self.children.append(child)
            if len(self.children):
                parent.children = still_parents_children
            parent.children.append(self)

    def is_root(self) -> bool:
        return self.parent is None and self.type_ is object

    @staticmethod
    def root() -> AbstractTypeHeap:
        return AbstractTypeHeap(object)

    def matches_exactly(self, type_in_question: type) -> bool | None:
        if self.type_ is type_in_question:
            return True
        if issubclass(type_in_question, self.type_):
            return False
        return None

    def find(
        self, type_in_question: type
    ) -> tuple[AbstractTypeHeap | None, AbstractTypeHeap | None]:
        assert self.is_root(), "it has to be called on root of the heap"
        m = self.matches_exactly(type_in_question)
        if m is True:
            return self, self.parent
        assert m is not None, f"it has to match root() or not to get here: {self}"
        return self._find_in_children(type_in_question)

    def _find_in_children(
        self, type_in_question: type
    ) -> tuple[AbstractTypeHeap | None, AbstractTypeHeap | None]:
        for child in self.children:
            m = child.matches_exactly(type_in_question)
            if m:
                return child, self
            if m is False:
                return child._find_in_children(type_in_question)
        return None, self

    def add(self, t: KnownType | KnownTypeArgs) -> None:
        found, found_parent = self.find(t.get_type())
        assert found is None, (
            f"Duplicate abstract type {t.get_type} but generally should not happen found= {found}, found_parent={found_parent}"
        )
        AbstractTypeHeap(t.get_type(), t, found_parent)

    @override
    def __repr__(self) -> str:
        return f"AbstractTypeHeap: {self.type_}, {len(self.children)}, {self.parent}"


class KnownTypesMap:
    types: dict[str, KnownType]
    types_by_type: dict[type, KnownType]
    abstract_types: AbstractTypeHeap

    def __init__(self) -> None:
        self.types = {}
        self.types_by_type = {}
        self.abstract_types = AbstractTypeHeap.root()

    def register(self, *array_of_args: KnownTypeArgs) -> None:
        for args in array_of_args:
            if not args.is_factory:
                self.register_type(KnownType(args))
            else:
                self.abstract_types.add(args)

    def register_type(self, t: KnownType) -> None:
        for alias in t.aliases:
            if alias not in self.types:
                self.types[alias] = t
        assert t.get_type() not in self.types_by_type, f"Duplicate type {t.get_type()}"
        self.types_by_type[t.type_] = t
        if t.is_abstract:
            self.abstract_types.add(t)

    def resolve_type(self, type_: type | str) -> KnownType:
        """
        >>> KNOWN_TYPES.resolve_type(int)
        KnownType('int', <class 'int'>)
        >>> KNOWN_TYPES.resolve_type("int")
        KnownType('int', <class 'int'>)
        >>> KNOWN_TYPES.resolve_type("zzz")
        Traceback (most recent call last):
        ...
        ValueError: Unknown type zzz
        >>> KNOWN_TYPES.resolve_type(dict)
        Traceback (most recent call last):
        ...
        ValueError: Unknown type <class 'dict'>
        >>> KNOWN_TYPES.resolve_type(3)
        Traceback (most recent call last):
        ...
        TypeError: issubclass() arg 1 must be a class
        """
        if isinstance(type_, str):
            type_ = type_.lower()
            if type_ in self.types:
                return self.types[type_]
        else:
            if type_ in self.types_by_type:
                return self.types_by_type[type_]
            found, super_type = self.abstract_types.find(type_)
            assert found is None, (
                f"Why it is not in types_by_type found={found}, super_type={super_type}"
            )
            if super_type is not None and not super_type.is_root():
                assert super_type.ktype is not None
                if isinstance(super_type.ktype, KnownTypeArgs):
                    args: KnownTypeArgs = super_type.ktype
                    ktype: KnownType = KnownType(args.build_concrete_type(type_))
                    self.register_type(ktype)
                    return ktype
                assert isinstance(super_type.ktype, KnownType), (
                    f"super_type.ktype is not a KnownType: {super_type.ktype}"
                )
                return super_type.ktype
        raise ValueError(f"Unknown type {type_}")


KNOWN_TYPES = KnownTypesMap()


KNOWN_TYPES.register(
    KnownTypeArgs(concrete_type=int),
    KnownTypeArgs(concrete_type=float),
    KnownTypeArgs(concrete_type=str, aliases=["literal"]),
    KnownTypeArgs(concrete_type=bool),
    KnownTypeArgs(concrete_type=bytes),
    KnownTypeArgs(
        concrete_type=date,
        map_to_string=lambda x: x.isoformat(),
        map_from_string=date.fromisoformat,
    ),
    KnownTypeArgs(
        concrete_type=datetime,
        map_to_string=lambda x: x.isoformat(),
        map_from_string=datetime.fromisoformat,
    ),
    KnownTypeArgs(abstract_type=Path, map_from_string=Path),
    KnownTypeArgs(
        abstract_type=Enum, is_factory=True, map_from_string=Enum, map_to_string=lambda x: x.value
    ),
    KnownTypeArgs(abstract_type=IntEnum, db_type=int, is_factory=True, map_from_db=IntEnum),
    KnownTypeArgs(
        abstract_type=BaseModel,
        json_type=dict,
        is_factory=True,
        map_from_json=BaseModel.model_validate,
        map_to_json=base_model_to_json,
    ),
    KnownTypeArgs(
        abstract_type=JsonBase,
        json_type=dict,
        map_from_json=JsonBase.from_json,
        map_to_json=base_model_to_json,
    ),
    # TODO: when appropriate packages are available entries below should be initialized
    # KnownType("dataframe", pd.DataFrame, json_type=dict),
)
